import sys
import click
import json

from taf import settings
from taf.api.repository import create_repository, taf_status
from taf.auth_repo import AuthenticationRepository
from taf.exceptions import TAFError, UpdateFailedError
from taf.log import initialize_logger_handlers, taf_logger
from taf.tools.cli import catch_cli_exception, find_repository
from taf.updater.types.update import UpdateType
from taf.updater.updater import OperationType, UpdateConfig, clone_repository, update_repository, validate_repository
from taf.yubikey.yubikey_manager import pin_managed


def common_update_options(f):
    f = click.option("--expected-repo-type", default="either", type=click.Choice(["test", "official", "either"]), help="Indicates expected authentication repository type - test or official.")(f)
    f = click.option("--scripts-root-dir", default=None, help="Scripts root directory, which can be used to move scripts out of the authentication repository for testing purposes.")(f)
    f = click.option("--profile", is_flag=True, help="Flag used to run profiler and generate .prof file")(f)
    f = click.option("--format-output", is_flag=True, help="Return formatted output which includes information on if build was successful and error message if it was raised")(f)
    f = click.option("--exclude-target", multiple=True, help="Globs defining which target repositories should be ignored during update.")(f)
    f = click.option("--strict", is_flag=True, default=False, help="Enable/disable strict mode - return an error if warnings are raised.")(f)
    return f


def _call_updater(config, format_output):
    """
    A helper function which calls update or clone repository
    """
    try:
        if config.operation == OperationType.CLONE:
            updater_output = clone_repository(config)
        else:
            updater_output = update_repository(config)

        successful = updater_output["event"] == "event/succeeded"
        if format_output:
            if successful:
                print(json.dumps({'updateSuccessful': successful}, ))
            else:
                error = updater_output.get("error_msg", "")
                print(json.dumps({'updateSuccessful': successful, "error": error}))

        if not successful:
            sys.exit(1)
    except Exception as e:
        if format_output:
            error_data = {"updateSuccessful": False, "error": str(e)}
            taf_logger.error(json.dumps(error_data))
            sys.exit(1)
        else:
            sys.exit(1)


def start_profiling():
    import cProfile
    import atexit

    print("Profiling...")
    pr = cProfile.Profile()
    pr.enable()

    def exit_profiler():
        pr.disable()
        print("Profiling completed")
        filename = 'updater.prof'  # You can change the filename if needed
        pr.dump_stats(filename)

    atexit.register(exit_profiler)


def create_repo_command():
    @click.command(help="""
        \b
        Create a new authentication repository at the specified location by registering
        signing keys and generating initial metadata files. Information about the roles
        can be provided through a dictionary - either specified directly or contained
        by a .json file whose path is specified when calling this command. This allows
        definition of:
            - total number of keys per role
            - threshold of signatures per role
            - should keys of a role be on Yubikeys or should keystore files be used
            - scheme (the default scheme is rsa-pkcs1v15-sha256)
            - keystore path, if not specified via keystore option

        \b
        For example:
        {
            "roles": {
                "root": {
                    "number": 3,
                    "length": 2048,
                    "passwords": ["password1", "password2", "password3"],
                    "threshold": 2,
                    "yubikey": true
                },
                "targets": {
                    "length": 2048
                },
                "snapshot": {},
                "timestamp": {}
            },
            "keystore": "keystore_path"
        }

        In cases when this dictionary is not specified, it is necessary to enter the needed
        information when asked to do so, or confirm that default values should be used.
        If keys should be stored in keystore files, it is possible to either use already generated
        keys (stored in keystore files located at the path specified using the keystore option),
        or to generate new one.

        If the test flag is set, a special target file will be created. This means that when
        calling the updater, it'll be necessary to use the --authenticate-test-repo flag.
        """)
    @catch_cli_exception(handle=TAFError, remove_dir_on_error=True)
    @click.argument("path", type=click.Path(exists=False, file_okay=False, dir_okay=True, writable=True))
    @click.option("--keys-description", help="A dictionary containing information about the "
                  "keys or a path to a json file which stores the needed information")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates if the changes should be "
                  "committed automatically")
    @click.option("--test", is_flag=True, default=False, help="Indicates if the created repository "
                  "is a test authentication repository")
    @pin_managed
    def create(path, keys_description, keystore, no_commit, test, pin_manager):
        create_repository(
            path=path,
            pin_manager=pin_manager,
            keystore=keystore,
            roles_key_infos=keys_description,
            commit=not no_commit,
            test=test,
        )
    return create


def clone_repo_command():
    @click.command(help="""
        Validate and clone authentication repositories and target repositories. URL of the
        remote authentication repository must be specified when calling this command. If the remote repository's URL is a file system path, the --from-fs flag must be used.

        The path to the authentication repository directory either read from targets/protected/info.json
        or specified using the --path option. If targets/protected/info.json does not exist and path is not
        defined, an error will be raised.

        If the authentication repository and the target repositories are in the same root directory,
        the locations of the target repositories are calculated based on the authentication repository's
        path. If this is not the case, it is necessary to redefine this default value using the
        --library-dir option. For example, if the authentication repository's path is
        E:\\root\\namespace\\auth-repo, and --library-dir is not specified, E:\\root is assumed
        to be the root directory.

        The names of the target repositories (as defined in repositories.json) are appended to the root
        repository's path, thus defining the location of each target repository. For instance, if the
        names of the target repositories are namespace/repo1, namespace/repo2, etc., and the root
        directory is E:\\root, the paths of the target repositories will be calculated as
        E:\\root\\namespace\\repo1, E:\\root\\namespace\\repo2, etc.

        The --scripts-root-dir option can be used to move scripts out of the authentication repository for
        testing purposes (to avoid a dirty index). Scripts are expected to be located in the
        scripts_root_dir/repo_name directory.

        One or more target repositories can be excluded from the update process using the --exclude-target
        option. In this case, the library will only be partly validated, so the last_validated_commit will
        not be updated, and no scripts will be called.

        The update can be performed in strict or non-strict mode. Strict mode is enabled by specifying
        --strict, which will raise errors during the update if any warnings are found. By default, --strict
        is disabled.
        """)
    @catch_cli_exception(handle=UpdateFailedError, skip_cleanup=True)
    @click.argument("url")
    @common_update_options
    @click.option("--path", help="Authentication repository's location. If not specified, calculated by combining repository's name specified in info.json and library dir")
    @click.option("--library-dir", default=None, help="Directory where target repositories and, optionally, authentication repository are located. If not specified, set to the current directory")
    @click.option("--from-fs", is_flag=True, default=False, help="Indicates if we want to clone a repository from the filesystem")
    @click.option("--bare", is_flag=True, default=False, help="Clone repositories as bare repositories")
    @click.option("--no-deps", is_flag=True, default=False, help="Optionally disables updating of dependencies")
    @click.option("--upstream/--no-upstream", default=False, help="Skips comparison with remote repositories upstream")
    @click.option("-v", "--verbosity", count=True, help="Displays varied levels of logging information based on verbosity level")
    @click.option("--run-scripts/--no-run-scripts", default=False, help="Run the auxiliary lifecycle handler scripts.")
    def clone(path, url, library_dir, from_fs, expected_repo_type, scripts_root_dir, profile, format_output, exclude_target, strict, bare, upstream, no_deps, verbosity, run_scripts):
        settings.VERBOSITY = verbosity
        initialize_logger_handlers()
        if profile:
            start_profiling()

        config = UpdateConfig(
            operation=OperationType.CLONE,
            remote_url=url,
            path=path,
            library_dir=library_dir,
            update_from_filesystem=from_fs,
            expected_repo_type=UpdateType(expected_repo_type),
            scripts_root_dir=scripts_root_dir,
            excluded_target_globs=exclude_target,
            strict=strict,
            bare=bare,
            no_upstream=not upstream,
            no_deps=no_deps,
            run_scripts=run_scripts,
        )

        _call_updater(config, format_output)

    return clone


def update_repo_command():
    @click.command(help="""
        Update and validate local authentication repositories and target repositories.

        If the authentication repository and the target repositories are in the same root directory,
        the locations of the target repositories are calculated based on the authentication repository's
        path. If this is not the case, it is necessary to redefine this default value using the
        --library-dir option. This means that if the authentication repository's path is
        E:\\root\\namespace\\auth-repo, and --library-dir is not specified, E:\\root is assumed to be
        the root directory.

        The names of the target repositories (as defined in repositories.json) are appended to the root
        repository's path, thus defining the location of each target repository. For example, if the names
        of the target repositories are namespace/repo1, namespace/repo2, etc., and the root directory is
        E:\\root, the path of the target repositories will be calculated as E:\\root\\namespace\\repo1,
        E:\\root\\namespace\\repo2, etc.

        The path to the authentication repository directory is set by the --path option. If this option is
        not provided, the current directory is used.

        The --scripts-root-dir option can be used to move scripts out of the authentication repository for
        testing purposes (to avoid a dirty index). Scripts will then be expected to be located in the
        scripts_root_dir/repo_name directory.

        One or more target repositories can be excluded from the update process using the --exclude-target
        option. In this case, the library will only be partly validated, so the last_validated_commit will
        not be updated, and no scripts will be called.

        The update can be performed in strict or non-strict mode. Strict mode is enabled by specifying
        --strict, which will raise errors during the update if any warnings are found. By default, --strict
        is disabled.
        """)
    @find_repository
    @catch_cli_exception(handle=UpdateFailedError, skip_cleanup=True)
    @common_update_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--library-dir", default=None, help="Directory where target repositories and, optionally, authentication repository are located. If not specified, calculated based on the authentication repository's path")
    @click.option("--force", is_flag=True, default=False, help="Force Update repositories")
    @click.option("--no-deps", is_flag=True, default=False, help="Optionally disables updating of dependencies.")
    @click.option("--upstream/--no-upstream", default=False, help="Skips comparison with remote repositories upstream")
    @click.option("-v", "--verbosity", count=True, help="Displays varied levels of logging information based on verbosity level")
    @click.option("--run-scripts/--no-run-scripts", default=False, help="Run the auxiliary lifecycle handler scripts.")
    def update(path, library_dir, expected_repo_type, scripts_root_dir, profile, format_output, exclude_target, strict, no_deps, force, upstream, verbosity, run_scripts):
        settings.VERBOSITY = verbosity
        initialize_logger_handlers()

        if profile:
            start_profiling()

        config = UpdateConfig(
            operation=OperationType.UPDATE,
            path=path,
            library_dir=library_dir,
            expected_repo_type=UpdateType(expected_repo_type),
            scripts_root_dir=scripts_root_dir,
            excluded_target_globs=exclude_target,
            strict=strict,
            force=force,
            no_upstream=not upstream,
            no_deps=no_deps,
            run_scripts=run_scripts,
        )

        _call_updater(config, format_output)
    return update


def validate_repo_command():
    @click.command(help="""
        Validates an authentication repository which is already on the file system
        and its target repositories (which are also expected to be on the file system).
        Does not clone repositories, fetch changes or merge commits.

        Validation can be in strict or no-strict mode. Strict mode is set by specifying --strict, which will raise errors
        during validate if any/all warnings are found. By default, --strict is disabled.
        """)
    @find_repository
    @catch_cli_exception(handle=UpdateFailedError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--library-dir", default=None, help="Directory where target repositories and, "
                  "optionally, authentication repository are located. If omitted it is "
                  "calculated based on authentication repository's path. "
                  "Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
    @click.option("--from-commit", default=None, help="First commit which should be validated.")
    @click.option("--from-latest", is_flag=True, default=False, help="Use the last validated commit as the starting point.")
    @click.option("--exclude-target", multiple=True, help="globs defining which target repositories should be "
                  "ignored during update.")
    @click.option("--strict", is_flag=True, default=False, help="Enable/disable strict mode - return an error"
                  "if warnings are raised")
    @click.option("--no-targets", is_flag=True, default=False, help="Skips target repository validation and validates only authentication repositories")
    @click.option("--no-deps", is_flag=True, default=False, help="Optionally disables updating of dependencies")
    @click.option("-v", "--verbosity", count=True, help="Displays varied levels of logging information based on verbosity level")
    @click.option("--profile", is_flag=True, help="Flag used to run profiler and generate .prof file")
    def validate(path, library_dir, from_commit, from_latest, exclude_target, strict, no_targets, no_deps, verbosity, profile):
        settings.VERBOSITY = verbosity
        initialize_logger_handlers()
        if profile:
            start_profiling()
        auth_repo = AuthenticationRepository(path=path)
        bare = auth_repo.is_bare_repository
        if from_latest:
            from_commit = auth_repo.last_validated_commit
        validate_repository(path, library_dir, from_commit, exclude_target, strict, bare, no_targets, no_deps)
    return validate


def latest_commit_and_branch_command():
    @click.command(help="""Fetch and print the last validated commit hash and the default branch.
        Integrated into the pre-push git hook""")
    @find_repository
    @catch_cli_exception
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    def latest_commit_and_branch(path):
        auth_repo = AuthenticationRepository(path=path)
        last_validated_commit = auth_repo.last_validated_commit or ""
        default_branch = auth_repo.default_branch
        print(f"{default_branch},{last_validated_commit}")
    return latest_commit_and_branch


def status_command():
    @click.command(help="Prints the whole state of the library, including authentication repositories and its dependencies.")
    @find_repository
    @catch_cli_exception
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--library-dir", default=None, help="Path to the library's root directory. Determined based on the authentication repository's path if not provided.")
    def status(path, library_dir):
        try:
            taf_status(path, library_dir)
        except TAFError as e:
            click.echo()
            click.echo(f"Error: {e}")
            click.echo()
    return status


def attach_to_group(group):
    group.add_command(create_repo_command(), name='create')
    group.add_command(clone_repo_command(), name='clone')
    group.add_command(update_repo_command(), name='update')
    group.add_command(validate_repo_command(), name='validate')
    group.add_command(latest_commit_and_branch_command(), name='latest-commit-and-branch')
    group.add_command(status_command(), name='status')



import click
import datetime
import os
from contextlib import contextmanager
from functools import wraps
from getpass import getpass
from pathlib import Path
from typing import Callable, Optional

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa, padding

from taf.api.utils._conf import find_taf_directory
from taf.tuf.keys import _get_legacy_keyid, get_sslib_key_from_value
from taf.yubikey.yubikey_manager import PinManager
from taf.models.types import KeysMapping
from ykman.device import list_all_devices
from yubikit.core.smartcard import SmartCardConnection
from ykman.piv import (
    KEY_TYPE,
    MANAGEMENT_KEY_TYPE,
    SLOT,
    PivSession,
    generate_random_management_key,
)
from yubikit.piv import (
    DEFAULT_MANAGEMENT_KEY,
    PIN_POLICY,
    InvalidPinError,
)

from taf.config import load_config
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import InvalidPINError, YubikeyError
from taf.utils import get_pin_for
from taf.log import taf_logger

from securesystemslib.signer._key import SSlibKey

DEFAULT_PIN = "123456"
DEFAULT_PUK = "12345678"
EXPIRATION_INTERVAL = 36500


def raise_yubikey_err(msg: Optional[str] = None) -> Callable:
    """Decorator used to catch all errors raised by yubikey-manager and raise
    YubikeyError. We don't need to handle specific cases.
    """

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except YubikeyError:
                raise
            except Exception as e:
                err_msg = (
                    f"{msg} Reason: ({type(e).__name__}) {str(e)}" if msg else str(e)
                )
                raise YubikeyError(err_msg) from e

        return decorator

    return wrapper


@contextmanager
def _yk_piv_ctrl(serial=None):
    """Context manager to open connection and instantiate Piv Session.

    Args:
        - serial (str): Match Yubikey's serial multiple keys are inserted

    Returns:
        - ykman.piv.PivSession

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Entering _yk_piv_ctrl context manager with serial={serial}")
    sessions = []
    devices_info = []
    try:
        for dev, info in list_all_devices():
            taf_logger.debug(f"Found device with info serial={info.serial}")
            if serial is None or info.serial == serial:
                connection = dev.open_connection(SmartCardConnection)
                try:
                    session = PivSession(connection)
                    sessions.append((session, info.serial))
                    devices_info.append((connection, session))
                    if serial is not None:
                        break
                except Exception as e:
                    connection.close()  # Ensure we close connection on error
                    raise e
        if serial is not None:
            session, serial = sessions[0]
            taf_logger.debug(f"Yielding single session for serial={serial}")
            yield [(session, serial)]
        else:
            taf_logger.debug(f"Yielding all sessions: {[s[1] for s in sessions]}")
            yield sessions
    finally:
        # Cleanup: ensure all connections are closed properly
        for connection, _ in devices_info:
            taf_logger.debug("Closing SmartCardConnection.")
            connection.close()


def is_inserted():
    """Checks if YubiKey is inserted.

    Args:
        None

    Returns:
        True if at least one Yubikey is inserted (bool)

    Raises:
        - YubikeyError
    """
    inserted_devices = list(list_all_devices())
    taf_logger.debug(f"is_inserted() sees {len(inserted_devices)} device(s)")
    return len(inserted_devices) > 0


@raise_yubikey_err()
def is_valid_pin(pin: str, serial: Optional[str] = None):
    """Checks if given pin is valid.

    Args:
        pin(str): Yubikey piv PIN

    Returns:
        tuple: True if PIN is valid, otherwise False, number of PIN retries

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Validating PIN for serial={serial}")
    if serial is None:
        serials = get_serial_nums()
        if len(serials) != 1:
            raise YubikeyError(
                "Please insert exactly one YubiKey or specify a serial number of the YubiKey whose pin is to be checked"
            )
        serial = serials[0]
    with _yk_piv_ctrl(serial=serial) as [(ctrl, _)]:
        try:
            ctrl.verify_pin(pin)
            taf_logger.debug(f"PIN is valid for serial={serial}")
            return True, None  # ctrl.get_pin_tries() fails if PIN is valid
        except InvalidPinError:
            tries_left = ctrl.get_pin_attempts()
            taf_logger.debug(
                f"Invalid PIN for serial={serial}, tries left = {tries_left}"
            )
            return False, tries_left


@raise_yubikey_err("Cannot get serial number.")
def get_serial_nums():
    """Get serial numbers of all inserted  YubiKeys

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Yubikey serial numbers

    Raises:
        - YubikeyError
    """
    serials = []
    with _yk_piv_ctrl() as sessions:
        for _, serial in sessions:
            taf_logger.debug(f"Discovered YubiKey with serial={serial}")
            serials.append(serial)
    return serials


@raise_yubikey_err("Cannot export x509 certificate.")
def export_piv_x509(cert_format=serialization.Encoding.PEM, serial=None):
    """Exports YubiKey's piv slot x509.

    Args:
        - cert_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV x509 certificate in a given format (bytes)

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Exporting X509 certificate for serial={serial}")
    with _yk_piv_ctrl(serial=serial) as [(ctrl, _)]:
        x509_cert = ctrl.get_certificate(SLOT.SIGNATURE)
        return x509_cert.public_bytes(encoding=cert_format)


@raise_yubikey_err("Cannot export public key.")
def export_piv_pub_key(pub_key_format=serialization.Encoding.PEM, serial=None):
    """Exports YubiKey's piv slot public key.

    Args:
        - pub_key_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV public key in a given format (bytes)

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Exporting public key for serial={serial}")
    with _yk_piv_ctrl(serial=serial) as [(ctrl, _)]:
        try:
            x509_cert = ctrl.get_certificate(SLOT.SIGNATURE)
            public_key = x509_cert.public_key()
            return public_key.public_bytes(
                encoding=pub_key_format,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        except Exception as e:
            raise YubikeyError(f"Failed to export public key: {str(e)}") from e


@raise_yubikey_err("Cannot export yk certificate.")
def export_yk_certificate(certs_dir, key: SSlibKey, serial: str):
    if certs_dir is None:
        certs_dir = Path.home()
    else:
        certs_dir = Path(certs_dir)
    certs_dir.mkdir(parents=True, exist_ok=True)
    cert_path = certs_dir / f"{key.keyid}.cert"
    print(f"Exporting certificate to {cert_path}")
    taf_logger.debug(f"Writing certificate file for keyid={key.keyid}, serial={serial}")
    with open(cert_path, "wb") as f:
        f.write(export_piv_x509(serial=serial))


def get_and_validate_pin(
    key_name,
    pin_confirm=True,
    pin_repeat=True,
    serial=None,
    key_id_pins=None,
    public_key=None,
):
    valid_pin = False

    if key_id_pins and public_key:
        key_id = _get_legacy_keyid(public_key)
        pin = key_id_pins.get(key_id)
    else:
        pin = None

    while not valid_pin:
        if not pin:
            pin = get_pin_for(key_name, pin_confirm, pin_repeat)

        valid_pin, retries = is_valid_pin(pin, serial)
        if not valid_pin and not retries:
            taf_logger.debug(f"PIN invalid and no retries left for serial={serial}")
            raise InvalidPINError("No retries left. YubiKey locked.")
        if not valid_pin:
            if not click.confirm(
                f"Incorrect PIN. Do you want to try again? {retries} retires left."
            ):
                taf_logger.debug("User cancelled PIN input.")
                raise InvalidPINError("PIN input cancelled")
            pin = None
    return pin


def get_pin_from_env(
    public_key: Optional[SSlibKey], serial_num: int, taf_dir: Optional[Path]
) -> Optional[str]:
    """Get PIN from environment variables.
    The PIN is stored in an environment variable named after either:

    i) the serial number of the YubiKey, or
    ii) the key name in uppercase, with hyphens replaced by underscores.

    For example, If the serial number is "123456", the environment variable would be "PIN_123456".
    Moreover, if the key name is "my-key", the environment variable
    would be "PIN_MY_KEY".
    If the key name is not found in the environment variables, the function
    returns None.
    """
    from taf.auth_repo import AuthenticationRepository

    pin = os.environ.get(f"PIN_{serial_num}")
    if pin is not None:
        return pin

    try:
        if taf_dir is None:
            raise FileNotFoundError(
                "No .taf directory found in the current working directory."
            )
        cfg = load_config(taf_dir / "config.toml")
        if cfg.root is None:
            raise FileNotFoundError("No root authentication repository found.")
        taf_logger.debug(f"Config loaded: {cfg}")
    except FileNotFoundError as e:
        taf_logger.debug(f"No config file found, skipping PIN from env. {str(e)}")
        return None

    if public_key is None:
        taf_logger.debug("No public key provided, skipping PIN from env.")
        return None
    root_auth_repo_name = cfg.root.name
    archive_dir = taf_dir.parent
    root_auth_repo = AuthenticationRepository(path=(archive_dir / root_auth_repo_name))
    if not root_auth_repo.is_git_repository:
        taf_logger.debug(
            f"{root_auth_repo_name} is not a valid authentication repository."
        )
        return None

    raw_mapping = root_auth_repo.get_keys_mapping()
    if raw_mapping is None:
        taf_logger.debug(
            f"No keys mapping found for root auth repo {root_auth_repo}, skipping PIN from env."
        )
        return None

    return _pin_from_keys_mapping(public_key, raw_mapping)


def _pin_from_keys_mapping(public_key: SSlibKey, mapping: dict) -> Optional[str]:
    """
    Return the PIN stored in PIN_<KEY_NAME>, where <KEY_NAME> is the entry in
    keys-mapping.json that matches `public_key`. If no match or no env var,
    return None.
    """
    key_name = KeysMapping.from_dict(mapping).find_name_by_public(
        public_key.keyval["public"]
    )
    if key_name:
        return os.getenv(f"PIN_{to_env_var_upper(key_name)}")
    return None


@raise_yubikey_err("Cannot get public key in TUF format.")
def get_piv_public_key_tuf(
    scheme=DEFAULT_RSA_SIGNATURE_SCHEME, serial=None
) -> SSlibKey:
    """Return public key from a Yubikey in TUF's RSAKEY_SCHEMA format.

    Args:
        - scheme(str): Rsa signature scheme (default is rsa-pkcs1v15-sha256)
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        A dictionary containing the RSA keys and other identifying information
        from inserted smart card.
        Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Extracting TUF-format public key from serial={serial}")
    pub_key_pem = export_piv_pub_key(serial=serial).decode("utf-8")
    return get_sslib_key_from_value(pub_key_pem, scheme)


def list_connected_yubikeys():
    """Lists all connected YubiKeys with their serial numbers and details."""
    yubikeys = list_all_devices()
    if not yubikeys:
        print("No YubiKeys connected.")
        taf_logger.debug("list_connected_yubikeys found no devices.")
    else:
        taf_logger.debug(f"list_connected_yubikeys found {len(yubikeys)} device(s).")
        for index, (_, info) in enumerate(yubikeys, start=1):
            print(f"YubiKey {index}:")
            print(f"  Serial Number: {info.serial}")
            print(f"  Version: {info.version}")
            print(f"  Form Factor: {info.form_factor}")


def _read_and_check_single_yubikey(
    role,
    key_name,
    taf_repo,
    pin_manager,
    registering_new_key,
    creating_new_key,
    pin_confirm,
    pin_repeat,
    prompt_message,
    retrying,
    yubikeys_to_skip,
    key_id_pins,
):
    taf_logger.debug(
        f"_read_and_check_single_yubikey for role='{role}', key_name='{key_name}', retrying={retrying}"
    )
    if retrying:
        if prompt_message is None:
            prompt_message = f"Please insert {key_name} YubiKey and press ENTER"
        getpass(prompt_message)

    if not yubikeys_to_skip:
        yubikeys_to_skip = []

    # make sure that YubiKey is inserted
    try:
        serials = get_serial_nums()
        taf_logger.debug(f"Found serials={serials} for single key check.")
        if taf_repo is None:
            # if setting up a YubiKey outside of the creation of a new repository or addition of new roles
            if len(serials) > 1:
                print("\nPlease insert only one YubiKey\n")
                taf_logger.debug("Multiple YubiKeys inserted, cannot proceed.")
                return
        else:
            not_loaded = [
                serial
                for serial in serials
                if not taf_repo.yubikey_store.is_loaded_for_role(serial, role)
                and serial not in yubikeys_to_skip
            ]
            taf_logger.debug(f"Not loaded for role='{role}': {not_loaded}")

            if len(not_loaded) != 1:
                print("\nPlease insert only one not previously inserted YubiKey\n")
                return None

            if not len(not_loaded):
                return None

            # no need to try loading keys that we know were previously loaded
            serials = not_loaded

    except Exception:
        taf_logger.log("NOTICE", "No YubiKeys inserted")
        return None

    serial_num = serials[0]
    # check if this key is already loaded as the provided role's key (we can use the same key
    # to sign different metadata)
    # read the public key, unless a new key needs to be generated on the yubikey
    public_key = (
        get_piv_public_key_tuf(serial=serial_num) if not creating_new_key else None
    )
    # check if this yubikey is can be used for signing the provided role's metadata
    # if the key was already registered as that role's key
    if not registering_new_key and role is not None and taf_repo is not None:
        if not taf_repo.is_valid_metadata_yubikey(role, public_key):
            taf_logger.debug(f"Public key not valid for role='{role}'.")
            return None

    pin = None
    if pin_manager.get_pin(serial_num) is None:
        if creating_new_key:
            pin = get_pin_for(key_name, pin_confirm, pin_repeat)
            taf_logger.debug("Attempting to load key pin from environment variables")

        taf_dir = find_taf_directory(Path().cwd())
        pin = get_pin_from_env(public_key, serial_num, taf_dir)

        if pin is None:
            pin = get_and_validate_pin(
                key_name,
                pin_confirm,
                pin_repeat,
                serial_num,
                key_id_pins=key_id_pins,
                public_key=public_key,
            )
        pin_manager.add_pin(serial_num, pin)

    if taf_repo is not None:
        # when reusing the same yubikey, public key will already be in the public keys dictionary
        # but the key name still needs to be added to the key id mapping dictionary
        taf_repo.yubikey_store.add_key_data(key_name, serial_num, public_key, role)

    taf_logger.debug(
        f"_read_and_check_single_yubikey returning for serial={serial_num}, key_name='{key_name}'"
    )
    return public_key, serial_num, key_name


def _read_and_check_yubikeys(
    role,
    taf_repo,
    pin_manager,
    pin_confirm,
    pin_repeat,
    prompt_message,
    key_names,
    retrying,
    hide_already_loaded_message,
    hide_threshold_message,
    key_id_pins,
):
    taf_logger.debug(
        f"_read_and_check_yubikeys for role='{role}', key_names={key_names}, retrying={retrying}"
    )
    if retrying:
        if prompt_message is None:
            if not hide_threshold_message:
                threshold = taf_repo.get_role_threshold(role)
                prompt_message = f"Please insert {role} ({', '.join(key_names)}) YubiKey(s) (threshold {threshold}) and press ENTER"
            else:
                prompt_message = f"Please insert {role} ({', '.join(key_names)}) YubiKey(s) and press ENTER"
        getpass(prompt_message)

    try:
        serials = get_serial_nums()
        taf_logger.debug(f"Detected YubiKey serials={serials}")
    except Exception:
        taf_logger.log("NOTICE", "No YubiKeys inserted")
        return None

    if not len(serials):
        taf_logger.debug("No YubiKeys present.")
        return None

    # check if this key is already loaded as the provided role's key (we can use the same key
    # to sign different metadata)
    yubikeys = []
    invalid_keys = []
    all_loaded = True

    taf_dir = find_taf_directory(Path().cwd())

    for index, serial_num in enumerate(serials):
        if not taf_repo.yubikey_store.is_loaded_for_role(serial_num, role):
            all_loaded = False
            # read the public key, unless a new key needs to be generated on the yubikey
            public_key = get_piv_public_key_tuf(serial=serial_num)
            # check if this yubikey is can be used for signing the provided role's metadata
            # if the key was already registered as that role's key
            if role is not None and taf_repo is not None:
                if not taf_repo.is_valid_metadata_yubikey(role, public_key):
                    invalid_keys.append(serial_num)
                    taf_logger.debug(
                        f"Serial={serial_num} not valid for role='{role}'."
                    )
                    continue

            key_name = taf_repo.keys_name_mappings.get(public_key.keyid)
            taf_logger.debug(
                f"Potential YubiKey with serial={serial_num}, associated key_name='{key_name}'."
            )
            pin = None
            if pin_manager.get_pin(serial_num) is None:
                pin = get_pin_from_env(public_key, serial_num, taf_dir)
                if pin is None:
                    pin = get_and_validate_pin(
                        key_name,
                        pin_confirm,
                        pin_repeat,
                        serial_num,
                        key_id_pins=key_id_pins,
                        public_key=public_key,
                    )
                pin_manager.add_pin(serial_num, pin)

            # when reusing the same yubikey, public key will already be in the public keys dictionary
            # but the key name still needs to be added to the key id mapping dictionary
            taf_repo.yubikey_store.add_key_data(key_name, serial_num, public_key, role)
            yubikeys.append((public_key, serial_num, key_name))

    if not hide_already_loaded_message and all_loaded:
        print("All inserted YubiKeys already loaded")
        taf_logger.debug(
            f"All YubiKeys inserted were already loaded for role='{role}'."
        )

    return yubikeys


@raise_yubikey_err("Cannot sign data.")
def sign_piv_rsa_pkcs1v15(data, pin, serial=None):
    """Sign data with key from YubiKey's piv slot.

    Args:
        - data(bytes): Data to be signed
        - pin(str): Pin for piv slot login.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Signature (bytes)

    Raises:
        - YubikeyError
    """
    taf_logger.debug(f"Signing data using YubiKey serial={serial}.")
    with _yk_piv_ctrl(serial=serial) as [(ctrl, _)]:
        ctrl.verify_pin(pin)
        sig = ctrl.sign(
            SLOT.SIGNATURE, KEY_TYPE.RSA2048, data, hashes.SHA256(), padding.PKCS1v15()
        )
        taf_logger.debug("Data signed successfully.")
        return sig


@raise_yubikey_err("Cannot setup Yubikey.")
def setup(
    pin,
    serial,
    cert_cn,
    cert_exp_days=365,
    pin_retries=10,
    private_key_pem=None,
    mgm_key=generate_random_management_key(MANAGEMENT_KEY_TYPE.TDES),
    key_size=2048,
):
    """Use to setup inserted Yubikey, with following steps (order is important):
      - reset to factory settings
      - set management key
      - generate key(RSA2048) or import given one
      - generate and import self-signed certificate(X509)
      - set pin retries
      - set pin
      - set puk(same as pin)

    Args:
        - cert_cn(str): x509 common name
        - cert_exp_days(int): x509 expiration (in days from now)
        - pin_retries(int): Number of retries for PIN
        - private_key_pem(str): Private key in PEM format. If given, it will be
                                imported to Yubikey.
        - mgm_key(bytes): New management key

    Returns:
        PIV public key in PEM format (bytes)

    Raises:
        - YubikeyError
    """
    taf_logger.debug(
        f"Initializing YubiKey setup for serial={serial}, key_size={key_size}, cert_cn='{cert_cn}'"
    )
    with _yk_piv_ctrl(serial=serial) as [(ctrl, _)]:
        taf_logger.debug(f"Resetting YubiKey to factory settings for serial={serial}")
        ctrl.reset()

        taf_logger.debug("Setting new management key and authenticating...")
        ctrl.authenticate(MANAGEMENT_KEY_TYPE.TDES, DEFAULT_MANAGEMENT_KEY)
        ctrl.set_management_key(MANAGEMENT_KEY_TYPE.TDES, mgm_key)

        if private_key_pem is None:
            taf_logger.debug("Generating RSA private key on the fly...")
            private_key = rsa.generate_private_key(65537, key_size, default_backend())
            pub_key = private_key.public_key()
        else:
            taf_logger.debug("Loading provided private key PEM...")
            try:
                private_key = load_pem_private_key(
                    private_key_pem, None, default_backend()
                )
            except TypeError:
                pem_pwd = getpass("Enter pem file password:\n")
                if pem_pwd:
                    pem_pwd = pem_pwd.encode()
                private_key = load_pem_private_key(
                    private_key_pem, pem_pwd, default_backend()
                )

        taf_logger.debug("Placing key in SIGNATURE slot with PIN_POLICY.ALWAYS...")
        ctrl.put_key(SLOT.SIGNATURE, private_key, PIN_POLICY.ALWAYS)
        pub_key = private_key.public_key()
        ctrl.authenticate(MANAGEMENT_KEY_TYPE.TDES, mgm_key)
        ctrl.verify_pin(DEFAULT_PIN)

        now = datetime.datetime.now()
        valid_to = now + datetime.timedelta(days=cert_exp_days)

        taf_logger.debug("Generating and importing self-signed certificate...")
        name = x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, cert_cn)])
        cert = (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(pub_key)
            .serial_number(x509.random_serial_number())
            .not_valid_before(now)
            .not_valid_after(valid_to)
            .sign(private_key, hashes.SHA256(), default_backend())
        )

        ctrl.put_certificate(SLOT.SIGNATURE, cert)
        taf_logger.debug("Setting PIN attempts and changing default PIN/PUK...")
        ctrl.set_pin_attempts(pin_attempts=pin_retries, puk_attempts=pin_retries)
        ctrl.change_pin(DEFAULT_PIN, pin)
        ctrl.change_puk(DEFAULT_PUK, pin)

    taf_logger.debug("YubiKey setup complete.")
    return pub_key.public_bytes(
        serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
    )


def setup_new_yubikey(
    pin_manager: PinManager,
    serial: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    key_size: Optional[int] = 2048,
) -> SSlibKey:
    taf_logger.debug(
        f"Starting new YubiKey setup for serial={serial}, scheme={scheme}, key_size={key_size}"
    )
    pin = pin_manager.get_pin(serial)
    cert_cn = input("Enter key holder's name: ")
    print("Generating key, please wait...")
    pub_key_pem = setup(
        pin, serial, cert_cn, cert_exp_days=EXPIRATION_INTERVAL, key_size=key_size
    ).decode("utf-8")
    scheme = DEFAULT_RSA_SIGNATURE_SCHEME
    key = get_sslib_key_from_value(pub_key_pem, scheme)
    taf_logger.debug(
        f"New YubiKey setup complete for serial={serial}, keyid={key.keyid}"
    )
    return key


def verify_yk_inserted(serial_num, key_name):
    def _check_if_yk_inserted():
        try:
            serials = get_serial_nums()
        except Exception:
            return False

        return serial_num in serials

    while not _check_if_yk_inserted():
        prompt_message = f"Please insert {key_name} YubiKey and press ENTER"
        taf_logger.debug(f"Waiting for user to insert YubiKey serial={serial_num}")
        getpass(prompt_message)


def yubikey_prompt(
    key_names,
    pin_manager,
    role=None,
    taf_repo=None,
    registering_new_key=False,
    creating_new_key=False,
    pin_confirm=True,
    pin_repeat=True,
    prompt_message=None,
    retry_on_failure=True,
    hide_already_loaded_message=False,
    hide_threshold_message=False,
    yubikeys_to_skip=None,
    key_id_pins=None,
):
    """Prompt for YubiKey insertion, handle reading, verifying, and returning the discovered YubiKey(s)."""
    taf_logger.debug(
        f"yubikey_prompt called with role='{role}', key_names={key_names}, registering_new_key={registering_new_key}, creating_new_key={creating_new_key}"
    )
    retry_counter = 0
    yubikeys = None

    if key_id_pins is None:
        key_id_pins = {}

    while True:
        retrying = retry_counter > 0
        taf_logger.debug(
            f"YubiKey prompt iteration={retry_counter}, retrying={retrying}"
        )

        if registering_new_key or creating_new_key:
            yubikey = _read_and_check_single_yubikey(
                role,
                key_names[0],
                taf_repo,
                pin_manager,
                registering_new_key,
                creating_new_key,
                pin_confirm,
                pin_repeat,
                prompt_message,
                retrying,
                yubikeys_to_skip,
                key_id_pins,
            )
            if yubikey:
                yubikeys = [yubikey]
        else:
            yubikeys = _read_and_check_yubikeys(
                role,
                taf_repo,
                pin_manager,
                pin_confirm,
                pin_repeat,
                prompt_message,
                key_names,
                retrying,
                hide_already_loaded_message,
                hide_threshold_message,
                key_id_pins,
            )

        if not yubikeys and not retry_on_failure:
            taf_logger.debug("No YubiKeys found and retry_on_failure=False. Returning.")
            return [(None, None, None)]
        if yubikeys:
            taf_logger.debug(f"Returning discovered YubiKeys: {yubikeys}")
            return yubikeys

        retry_counter += 1


def yk_secrets_handler(prompt, pin_manager, serial_num):
    taf_logger.debug(
        f"yk_secrets_handler called with prompt='{prompt}', serial_num={serial_num}"
    )
    if prompt == "pin":
        return pin_manager.get_pin(serial_num)
    raise YubikeyError(f"Invalid prompt {prompt}")


def to_env_var_upper(key_name):
    """Convert key name to uppercase and replace '-' with '_' for environment variable."""
    return key_name.upper().replace("-", "_")



import click
from taf.tools.cli.lazy_group import LazyGroup


def taf_command_loader(name):
    if name != "yubikey":
        if name in (
            "dependencies",
            "keystore",
            "conf",
            "targets",
            "metadata",
            "roles",
            "repo",
        ):
            return f"taf.tools.{name}.attach_to_group"
        else:
            raise RuntimeError(f"Unknown command: {name}")
    try:
        import ykman  # noqa: F401

        return "taf.tools.yubikey.attach_to_group"
    except ImportError:
        pass


@click.group(
    cls=LazyGroup,
    lazy_subcommands={
        "dependencies": lambda: taf_command_loader("dependencies"),
        "keystore": lambda: taf_command_loader("keystore"),
        "conf": lambda: taf_command_loader("conf"),
        "repo": lambda: taf_command_loader("repo"),
        "targets": lambda: taf_command_loader("targets"),
        "metadata": lambda: taf_command_loader("metadata"),
        "roles": lambda: taf_command_loader("roles"),
        "yubikey": lambda: taf_command_loader("yubikey"),
    },
)
@click.version_option()
def taf():
    """TAF Command Line Interface"""
    pass


if __name__ == "__main__":
    taf()



from typing import Optional


class TAFKey:
    def __init__(
        self,
        key_id: str,
        name: Optional[str] = None,
        organization: Optional[str] = None,
        country: Optional[str] = None,
        state: Optional[str] = None,
        locality: Optional[str] = None,
        valid_from: Optional[str] = None,
        valid_to: Optional[str] = None,
    ):
        self.key_id = key_id
        self.name = name or ""
        self.organization = organization or ""
        self.country = country or ""
        self.state = state or ""
        self.locality = locality or ""
        self.valid_from = valid_from or ""
        self.valid_to = valid_to or ""

    def __str__(self):
        attributes = [
            f"Key ID: {self.key_id}",
            f"Name: {self.name}" if self.name else None,
            f"Organization: {self.organization}" if self.organization else None,
            f"Country: {self.country}" if self.country else None,
            f"State: {self.state}" if self.state else None,
            f"Locality: {self.locality}" if self.locality else None,
            f"Valid From: {self.valid_from}" if self.valid_from else None,
            f"Valid To: {self.valid_to}" if self.valid_to else None,
        ]

        return "\n".join(filter(None, attributes))



from shutil import Error, copytree
import shutil
from typing import Optional
from pathlib import Path
from taf.api.keystore import generate_keys
from taf.log import taf_logger
from taf.utils import read_input_dict


def init(
    path: Optional[str] = None,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
):
    # Determine the directory path
    if path:
        taf_directory = Path(path) / ".taf"
    else:
        taf_directory = Path(".taf")

    if taf_directory.exists() and taf_directory.is_dir():
        taf_logger.log("NOTICE", ".taf directory already exists.")
    else:
        # Create the .taf directory
        taf_directory.mkdir(exist_ok=True)
        taf_logger.log("NOTICE", "Generated .taf directory")

    # Create the config.toml file
    config_file_path = taf_directory / "config.toml"
    config_file_path.touch()  # Create an empty file

    # Create the keystore directory
    keystore_directory = taf_directory / "keystore"
    keystore_directory.mkdir(exist_ok=True)

    # If any of these parameters exist you can assume the user wants to generate keys

    # check if keystore already exists
    roles_key_infos_dict = read_input_dict(roles_key_infos)
    keystore = (
        keystore or (roles_key_infos and roles_key_infos_dict.get("keystore")) or None
    )
    should_generate_keys = False
    keystore_path = Path(keystore) if keystore else None
    if not keystore:
        # Prompt the user if they want to run the generate_keys function
        while True:
            use_keystore = (
                input("Do you want to generate keys in the keystore directory? [y/N]: ")
                .strip()
                .lower()
            )
            if use_keystore in ["y", "n"]:
                should_generate_keys = use_keystore == "y"
                break

        if should_generate_keys:
            # First check if the user already specified keystore
            copy_keystore = (
                input(
                    "Do you want to load an existing keystore from another location? [y/N]: "
                )
                .strip()
                .lower()
            )
            if copy_keystore == "y":
                while True:
                    keystore_input = input(
                        "Enter the path to the existing keystore:"
                    ).strip()
                    keystore_path = Path(keystore_input)
                    if keystore_path.exists() and keystore_path.is_dir():
                        keystore = keystore_input  # Assign the string path to the keystore variable
                        should_generate_keys = (
                            False  # no need to generate keys, they will be copied
                        )
                        break
                    else:
                        taf_logger.error(
                            f"Provided keystore path {keystore} is invalid."
                        )
    # Check if keystore is specified now. If so copy the keys
    if keystore and keystore_path and keystore_path.is_dir():
        try:
            copytree(keystore, keystore_directory, dirs_exist_ok=True)
            taf_logger.log(
                "NOTICE", f"Copied keystore from {keystore} to {keystore_directory}"
            )
        except FileNotFoundError:
            taf_logger.error(f"Provided keystore path {keystore} not found.")
        except Error as e:
            taf_logger.error(f"Error occurred while copying keystore: {e}")

    if should_generate_keys:
        generate_keys(keystore_directory, roles_key_infos)
        taf_logger.log(
            "NOTICE",
            f"Successfully generated keys inside the {keystore_directory} directory",
        )

    if roles_key_infos is not None and Path(roles_key_infos).is_file():
        infos_config_path = (taf_directory / Path(roles_key_infos).name).absolute()
        shutil.copy(str(roles_key_infos), str(infos_config_path))



from collections import defaultdict
import contextlib
from typing import Dict, List, Optional, Tuple
from taf.tuf.keys import SSlibKey


class YubiKeyStore:
    def __init__(self):
        # Initializes the dictionary to store YubiKey data
        self._yubikeys_data = defaultdict(dict)

    @property
    def yubikeys_data(self) -> Dict:
        return self._yubikeys_data

    def is_loaded(self, serial_number) -> bool:
        return any(
            data["serial"] == serial_number for data in self._yubikeys_data.values()
        )

    def is_loaded_for_role(self, serial_number: str, role_name: str) -> bool:
        for data in self._yubikeys_data.values():
            if data["serial"] == serial_number and role_name in data["roles"]:
                return True
        return False

    def is_key_name_loaded(self, key_name: str) -> bool:
        """Check if the key name is already loaded."""
        return key_name in self._yubikeys_data

    def add_key_data(
        self,
        key_name: str,
        serial_num: str,
        public_key: SSlibKey,
        role_name: str,
    ) -> None:
        """Add data associated with a YubiKey."""
        if role_name in self._yubikeys_data:
            key_data = self._yubikeys_data[key_name]
        else:
            key_data = {"serial": serial_num, "public_key": public_key, "roles": []}
        key_data["roles"].append(role_name)
        self._yubikeys_data[key_name] = key_data

    def get_key_data(self, key_name: str) -> Optional[Tuple[str, SSlibKey]]:
        """Retrieve data associated with a given YubiKey name."""
        if not self.is_key_name_loaded(key_name):
            return None
        key_data = self._yubikeys_data.get(key_name)
        return key_data["public_key"], key_data["serial"]

    def get_roles_of_key(self, serial_number: str) -> List[str]:
        roles = []
        for data in self._yubikeys_data.values():
            if data["serial"] == serial_number:
                roles.extend(data["roles"])
        return roles

    def remove_key_data(self, key_name: str) -> bool:
        """Remove data associated with a given YubiKey name if it exists."""
        if self.is_key_name_loaded(key_name):
            del self._yubikeys_data[key_name]
            return True
        return False


class PinManager:
    """
    Manages PIN storage and retrieval for YubiKeys, providing a centralized mechanism
    to handle adding and retrieving PINs to and from storage.

    Attributes:
        auto_continue (bool): If set to True, suppresses prompts when they are not necessary, allowing
                              uninterrupted execution. Default is False.
        _pins (dict): A private dictionary that stores PINs, keyed by an identifier (e.g., the YubiKey serial number).
    """

    def __init__(self, auto_continue=False):
        self._pins = {}
        # Automatically continue without prompts e.g. when attempting to load more keys
        self.auto_continue = auto_continue

    def add_pin(self, serial_number, pin):
        self._pins[serial_number] = pin

    def clear_pins(self):
        for key in list(self._pins.keys()):
            self._pins[key] = None
        self._pins.clear()

    def get_pin(self, serial_number):
        return self._pins.get(serial_number)


@contextlib.contextmanager
def manage_pins():
    """
    Context manager to handle the lifecycle of PinManager instances.

    Yields:
        PinManager: An instance of PinManager, ready for use.

    Ensures that PINs are cleared from memory immediately after the operations
    involving the PinManager are completed, maintaining security.
    """
    pin_manager = PinManager()
    try:
        yield pin_manager
    finally:
        pin_manager.clear_pins()


def pin_managed(func):
    """
    Decorator that wraps a function to automatically provide it with a managed PinManager instance.

    Args:
        func (Callable): The function to be decorated, which will use a PinManager.

    Returns:
        Callable: The wrapped function with a PinManager instance passed as a keyword argument.

    This decorator abstracts away the creation and cleanup of PinManager instances, allowing
    functions to utilize a PinManager without concern for its lifecycle.
    """

    def wrapper(*args, **kwargs):
        with manage_pins() as pin_manager:
            kwargs["pin_manager"] = pin_manager
            return func(*args, **kwargs)

    return wrapper



from functools import partial
from logging import INFO
from typing import Dict, List, Optional, Tuple, Union
import click
from pathlib import Path
from logdecorator import log_on_start
from taf.auth_repo import AuthenticationRepository
from taf.log import taf_logger
from taf.models.types import Role, RolesIterator
from taf.models.models import TAFKey
from taf.models.types import TargetsRole, MainRoles, UserKeyData
from taf.tuf.repository import MetadataRepository as TUFRepository, is_auth_repo
from taf.api.utils._conf import find_keystore
from taf.utils import secret_yes_no_prompt
from taf.tuf.keys import (
    YkSigner,
    _get_legacy_keyid,
    generate_and_write_rsa_keypair,
    generate_rsa_keypair,
    get_sslib_key_from_value,
    load_signer_from_pem,
)

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME, RoleSetupParams
from taf.exceptions import (
    KeystoreError,
    SigningError,
    YubikeyError,
)
from taf.keystore import (
    get_keystore_keys_of_role,
    key_cmd_prompt,
    load_signer_from_private_keystore,
)
from taf import YubikeyMissingLibrary
from securesystemslib.signer._crypto_signer import CryptoSigner

try:
    import taf.yubikey.yubikey as yk
except ImportError:
    taf_logger.warning(
        "WARNING: yubikey-manager dependency not installed. You will not be able to use YubiKeys."
    )
    yk = YubikeyMissingLibrary()  # type: ignore


def _create_signer(auth_repo, public_key, serial_num, key_name):
    return YkSigner(
        public_key,
        serial_num,
        partial(
            yk.yk_secrets_handler,
            pin_manager=auth_repo.pin_manager,
            serial_num=serial_num,
        ),
        key_name=key_name,
    )


def get_key_name(role_name: str, key_num: int, num_of_keys: int) -> str:
    """
    Return a keystore key's name based on the role's name and total number of signing keys,
    as well as the specified counter. If number of signing keys is one, return the role's name.
    If the number of signing keys is greater that one, return role's name + counter (root1, root2...)
    """
    if num_of_keys == 1:
        return role_name
    else:
        return role_name + str(key_num + 1)


def get_metadata_key_info(certs_dir: str, key_id: str) -> TAFKey:
    """
    Read and return information about the specified key read from a certificate
    file whose name matches that key's id.
    """
    cert_path = Path(certs_dir, key_id + ".cert")
    if cert_path.is_file():
        cert_pem = cert_path.read_bytes()
        return TAFKey(key_id, **_extract_x509(cert_pem))

    return TAFKey(key_id)


def _extract_x509(cert_pem: bytes) -> Dict:
    from cryptography import x509
    from cryptography.hazmat.backends import default_backend

    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())

    def _get_attr(oid):
        attrs = cert.subject.get_attributes_for_oid(oid)
        return attrs[0].value if len(attrs) > 0 else ""

    return {
        "name": _get_attr(x509.OID_COMMON_NAME),
        "organization": _get_attr(x509.OID_ORGANIZATION_NAME),
        "country": _get_attr(x509.OID_COUNTRY_NAME),
        "state": _get_attr(x509.OID_STATE_OR_PROVINCE_NAME),
        "locality": _get_attr(x509.OID_LOCALITY_NAME),
        "valid_from": cert.not_valid_before.strftime("%Y-%m-%d"),
        "valid_to": cert.not_valid_after.strftime("%Y-%m-%d"),
    }


def load_sorted_keys_of_new_roles(
    roles: Union[MainRoles, TargetsRole],
    auth_repo: AuthenticationRepository,
    yubikeys_data: Optional[Dict[str, UserKeyData]],
    keystore: Optional[Union[Path, str]],
    existing_roles: Optional[List[str]] = None,
    skip_prompt: Optional[bool] = False,
    certs_dir: Optional[Union[Path, str]] = None,
):
    """
    Load signing keys of roles - first those stored on YubiKeys to avoid entering pins
    if there is something wrong with keystore files, then from keystore files.
    Recursively load keys of all delegated roles of target roles.

    Arguments:
        auth_repo: Authentication repository's instance
        roles: MainRoles object (including root, targets, snapshot and timestamp) or a particular delegated role
        yubikeys_data: Contains a mapping of a YubiKey's name to additional details. There names
            are used to link roles to YubiKeys that are used to sign the corresponding metadata.
            If additional details contain the public key, a user will not have to insert that YubiKey
            (provided that it's not necessary given the threshold of signing keys)
        keystore: keystore path
        pint_manager: Instance of a class for secure pin management
        existing_roles (optional): A list of roles whose keys were already loaded
        skip_prompt (optional): A flag defining if the user will be asked if they want to generate new keys or reuse existing
            ones in case keystore files should be used. New keys will be generated by default.
    Side Effects:
        Populates loaded YubiKeys database located in `yubikey.py`

    Returns:
        Signing and verification keys of roles
    """

    def _sort_roles(roles):
        keystore_roles = []
        yubikey_roles = []
        for role in RolesIterator(roles):
            if role.is_yubikey:
                yubikey_roles.append(role)
            else:
                keystore_roles.append(role)
        return keystore_roles, yubikey_roles

    # load and/or generate all keys first
    if existing_roles is None:
        existing_roles = []
    try:
        keystore_roles, yubikey_roles = _sort_roles(roles)
        signers: Dict = {}
        verification_keys: Dict = {}

        for role in keystore_roles:
            if role.name in existing_roles:
                continue
            keystore_signers, _, _ = setup_roles_keys(
                role,
                auth_repo,
                keystore=keystore,
                skip_prompt=skip_prompt,
            )
            for signer in keystore_signers:
                signers.setdefault(role.name, []).append(signer)

        for role in yubikey_roles:
            if role.name in existing_roles:
                continue
            _, yubikey_keys, yubikey_signers = setup_roles_keys(
                role,
                auth_repo,
                certs_dir=certs_dir,
                users_yubikeys_details=yubikeys_data,
                skip_prompt=skip_prompt,
            )
            verification_keys[role.name] = yubikey_keys
            signers[role.name] = yubikey_signers

        return signers, verification_keys
    except KeystoreError:
        raise SigningError("Could not load keys of new roles")


def _load_signer_from_keystore(
    taf_repo, keystore_path, key_name, num_of_signatures, scheme, role
) -> CryptoSigner:
    if keystore_path is None:
        return None
    if (keystore_path / key_name).is_file():
        try:
            signer = load_signer_from_private_keystore(
                keystore=keystore_path, key_name=key_name, scheme=scheme
            )
            # load only valid keys
            if taf_repo.is_valid_metadata_key(role, signer.public_key, scheme=scheme):
                return signer
        except KeystoreError:
            pass

    return None


def _load_yubikeys(
    taf_repo: TUFRepository,
    role: str,
    key_names: List[str],
    retry_on_failure: bool,
    hide_threshold_message: bool,
    key_id_pins: Optional[Dict] = None,
):
    """
    Loads YubiKeys for a specified role using given key names and manages the interaction process.

    Returns:
        A tuple containing two elements:
            - list of YkSigner instances representing the signers initialized from the loaded YubiKeys.
            - list of key names that were successfully loaded.
    """
    taf_logger.debug(
        f"Attempting to load YubiKeys for role '{role}' with key names: {key_names}"
    )

    signers_yubikeys: List = []
    yubikeys = yk.yubikey_prompt(
        key_names=key_names,
        pin_manager=taf_repo.pin_manager,
        role=role,
        taf_repo=taf_repo,
        retry_on_failure=retry_on_failure,
        hide_already_loaded_message=True,
        hide_threshold_message=hide_threshold_message,
        key_id_pins=key_id_pins,
    )
    taf_logger.debug(
        f"yubikey_prompt returned: {yubikeys if yubikeys else 'No YubiKeys loaded'}"
    )

    loaded_keyids = [signer.public_key.keyid for signer in signers_yubikeys]
    taf_logger.debug(f"Currently loaded keyids for {role}: {loaded_keyids}")

    loaded_key_names = []
    for public_key, serial_num, key_name in yubikeys:
        if public_key is not None and public_key.keyid not in loaded_keyids:
            taf_logger.debug(
                f"Found YubiKey: Serial={serial_num}, key_name='{key_name}', keyid={public_key.keyid}"
            )
            signer = YkSigner(
                public_key,
                serial_num,
                partial(
                    yk.yk_secrets_handler,
                    pin_manager=taf_repo.pin_manager,
                    serial_num=serial_num,
                ),
                key_name=key_name,
            )
            signers_yubikeys.append(signer)
            loaded_keyids.append(public_key.keyid)
            loaded_key_names.append(key_name)
            taf_logger.info(f"Successfully loaded {key_name} from inserted YubiKey")

    return signers_yubikeys, loaded_key_names


@log_on_start(INFO, "Loading signing keys of '{role:s}'", logger=taf_logger)
def load_signers(
    taf_repo: TUFRepository,
    role: str,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    key_id_pins: Optional[Dict] = None,
    use_yubikeys_to_sign: bool = False,
) -> Tuple[List[Dict], List[Dict]]:
    """
    Load role's signing keys. Make sure that at least the threshold of keys was
    loaded, but allow loading more keys (so that a metadata file can be signed
    by all of the role's keys if the user wants that)
    """
    threshold = taf_repo.get_role_threshold(role)
    signing_keys_num = len(taf_repo.get_role_keys(role))
    all_loaded = False
    num_of_signatures = 0
    signers_keystore: List = []
    signers_yubikeys: List = []

    # first try to sign using yubikey
    # if that is not possible, try to load key from a keystore file
    # if the keystore file is not found, ask the user if they want to sign
    # using yubikey and to insert it if that is the case

    if keystore is None:
        keystore_path = find_keystore(taf_repo.path)
        if keystore_path is None:
            taf_logger.warning("No keystore provided and no default keystore found")
        else:
            keystore = str(keystore_path)

    keystore_path = Path(keystore).expanduser().resolve() if keystore else None

    keystore_files = []
    if keystore is not None:
        keystore_files = get_keystore_keys_of_role(keystore, role)
    prompt_for_yubikey = True

    taf_repo.add_default_names_of_role(role)
    key_names = taf_repo.get_key_names_of_role(role)
    initial_yk_load_attempt = True
    hide_threshold_message = False

    while not all_loaded and num_of_signatures < signing_keys_num:

        # when loading from keystore files
        # there is no need to ask the user if they want to load more key, try to load from keystore
        if num_of_signatures < len(keystore_files):
            key_name = keystore_files[num_of_signatures]
            signer = _load_signer_from_keystore(
                taf_repo, keystore_path, key_name, num_of_signatures, scheme, role
            )
            if signer is not None:
                signers_keystore.append(signer)
                num_of_signatures += 1
                continue
        if num_of_signatures >= threshold:
            if use_yubikeys_to_sign and not taf_repo.pin_manager.auto_continue:
                if not secret_yes_no_prompt(
                    f"Threshold of {role} keys reached. "
                    f"Do you want to load more {role} keys? [y/N]"
                ):
                    break
                else:
                    hide_threshold_message = True
            else:
                # loading from keystore files, couldn't load from all of them, but loaded enough
                # or auto continue sets
                break

        # try to load from the inserted YubiKeys, without asking the user to insert it
        # in case of yubikeys, instead of asking for a particular key, ask to insert all
        # that can be used to sign the current role, but either read the name from the
        # metadata, or assign a role + counter name
        if initial_yk_load_attempt or use_yubikeys_to_sign:
            loaded_signers, loaded_keys = _load_yubikeys(
                taf_repo=taf_repo,
                role=role,
                key_names=key_names,
                retry_on_failure=use_yubikeys_to_sign,
                hide_threshold_message=hide_threshold_message,
                key_id_pins=key_id_pins,
            )
            signers_yubikeys.extend(loaded_signers)
            if loaded_keys:
                use_yubikeys_to_sign = True
                num_of_loaded_keys = len(loaded_keys)
                for loaded_key in loaded_keys:
                    key_names.remove(loaded_key)

                if num_of_loaded_keys:
                    num_of_signatures += num_of_loaded_keys
                    continue

        if prompt_for_yubikey and not use_yubikeys_to_sign:
            if click.confirm(f"Sign {role} using YubiKey(s)?"):
                use_yubikeys_to_sign = True
                prompt_for_yubikey = False
                continue

        if prompt_for_keys and click.confirm(f"Manually enter {role} key?"):
            keys = [signer.public_key for signer in signers_keystore]
            key = key_cmd_prompt(key_name, role, taf_repo, keys, scheme)
            signer = load_signer_from_pem(key)
            signers_keystore.append(key)
            num_of_signatures += 1
        else:
            raise SigningError(f"Cannot load keys of role {role}")

    return signers_keystore, signers_yubikeys


def setup_roles_keys(
    role: Role,
    auth_repo: AuthenticationRepository,
    certs_dir: Optional[Union[Path, str]] = None,
    keystore: Optional[Union[Path, str]] = None,
    users_yubikeys_details: Optional[Dict[str, UserKeyData]] = None,
    skip_prompt: Optional[bool] = False,
    key_size: int = 2048,
):

    if role.name is None:
        raise SigningError("Cannot set up roles keys. Role name not specified")
    yubikey_keys = []
    keystore_signers = []
    yubikey_signers = []

    yubikey_ids = role.yubikey_ids
    if yubikey_ids is None:
        yubikey_ids = []
    if users_yubikeys_details is None:
        users_yubikeys_details = {}

    if yubikey_ids:
        if auth_repo.keys_name_mappings:
            # check if some of the listed key names are already defined as signing keys
            # in that case, they need to be loaded and verified
            if is_auth_repo(auth_repo.path):
                existing_key_names = {
                    existing_key_name: existing_key_id
                    for existing_key_id, existing_key_name in auth_repo.keys_name_mappings.items()
                }
                for key_name in yubikey_ids:
                    if key_name in existing_key_names:
                        public_key_pem, scheme = auth_repo.get_public_key_of_keyid(
                            existing_key_names[key_name]
                        )
                        key_data = {"public": public_key_pem, "scheme": scheme}
                        users_yubikeys_details[key_name] = UserKeyData(**key_data)

    if role.is_yubikey:
        taf_logger.debug(
            f"Setting up YubiKey-based role '{role.name}' requiring {role.number} key(s)."
        )
        yubikey_keys, yubikey_signers = _setup_yubikey_roles_keys(
            auth_repo, yubikey_ids, users_yubikeys_details, role, certs_dir, key_size
        )
    else:
        if keystore is None:
            taf_logger.error("No keystore provided and no default keystore found")
            raise KeyError("No keystore provided and no default keystore found")
        default_params = RoleSetupParams()
        for key_num in range(role.number):
            key_name = get_key_name(role.name, key_num, role.number)
            signer, key_id = _setup_keystore_key(
                keystore,
                role.name,
                key_name,
                role.scheme or default_params["scheme"],
                role.length or default_params["length"],
                None,
                skip_prompt=skip_prompt,
            )
            keystore_signers.append(signer)
            auth_repo.add_key_name(key_name, key_id)

    return keystore_signers, yubikey_keys, yubikey_signers


def _setup_yubikey_roles_keys(
    auth_repo, yubikey_ids, users_yubikeys_details, role, certs_dir, key_size
):
    loaded_keys_num = 0
    yk_with_public_key = {}
    yubikey_keys = []
    signers = []

    # if a key was already loaded (while setting up a different role)
    # register signers and remove the key id, so that the user is not asked to enter it again
    yubikes_to_skip = []
    names_defined = bool(yubikey_ids)
    if names_defined:
        for key_name in list(yubikey_ids):
            key_data = auth_repo.yubikey_store.get_key_data(key_name)
            if key_data is not None:
                public_key, serial_num = key_data
                auth_repo.yubikey_store.add_key_data(
                    key_name, serial_num, public_key, role.name
                )
                yubikey_ids.remove(key_name)
                yubikey_keys.append(public_key)
                loaded_keys_num += 1
                signer = _create_signer(auth_repo, public_key, serial_num, key_name)
                signers.append(signer)

        # if key already loaded while setting up a different role, skip it
        # if the current role's yubikey ids are defined
        # however, if the current role's yubikey ids are not specified
        # it can be possible to reuse a yubikey
        for key_name, key_data in auth_repo.yubikey_store.yubikeys_data.items():
            if key_name not in yubikey_ids:
                yubikes_to_skip.append(key_data["serial"])
    else:
        yubikey_ids = [f"{role.name}{counter}" for counter in range(1, role.number + 1)]

    for key_name in yubikey_ids:
        public_key_text = None
        if key_name in users_yubikeys_details:
            public_key_text = users_yubikeys_details[key_name].public
        if public_key_text:
            scheme = users_yubikeys_details[key_name].scheme
            public_key = get_sslib_key_from_value(public_key_text, scheme)
            # Check if the signing key is already loaded
            if not auth_repo.yubikey_store.is_key_name_loaded(key_name):
                yk_with_public_key[key_name] = public_key
            else:
                serial_num = auth_repo.yubikey_store.get_key_data["serial"]
                auth_repo.yubikey_store.add_key_data(
                    key_name, serial_num, public_key, role.name
                )
                loaded_keys_num += 1
            yubikey_keys.append(public_key)
        else:
            key_scheme = None
            if key_name in users_yubikeys_details:
                key_scheme = users_yubikeys_details[key_name].scheme
            key_scheme = key_scheme or role.scheme
            taf_logger.debug(
                f"Preparing to set up YubiKey for role '{role.name}', key_name='{key_name}'."
            )
            public_key, serial_num = _setup_yubikey(
                auth_repo,
                role.name,
                key_name,
                key_scheme,
                certs_dir,
                key_size,
                yubikes_to_skip,
            )
            loaded_keys_num += 1
            signer = _create_signer(auth_repo, public_key, serial_num, key_name)
            signers.append(signer)

        key_id = _get_legacy_keyid(public_key)
        auth_repo.add_key_name(key_name, key_id, overwrite=names_defined)

    if loaded_keys_num < role.number:
        if loaded_keys_num < role.threshold:
            print(f"Threshold of role {role.name} is {role.threshold}")
        taf_logger.debug(
            f"Only loaded {loaded_keys_num} out of {role.number} needed key(s) for '{role.name}'. Checking if threshold is met..."
        )

        _load_remaining_keys_of_role(
            auth_repo,
            role,
            loaded_keys_num,
            users_yubikeys_details,
            yk_with_public_key,
            yubikey_keys,
            signers,
        )
    return yubikey_keys, signers


def _setup_keystore_key(
    keystore: Optional[Union[Path, str]],
    role_name: str,
    key_name: str,
    scheme: str,
    length: int,
    password: Optional[str],
    skip_prompt: Optional[bool],
) -> Tuple[CryptoSigner, str]:
    generate_new_keys = keystore is None
    signer = None

    def _invalid_key_message(key_name, keystore):
        key_path = Path(keystore, key_name)
        if key_path.is_file():
            print(f"Could not load private key {key_path}")
        else:
            print(f"{key_path} is not a file!")

    if keystore is not None:
        keystore_path = str(Path(keystore).expanduser().resolve())
        while signer is None:
            try:
                signer = load_signer_from_private_keystore(
                    keystore_path,
                    key_name,
                    scheme=scheme,
                    password=password,
                )
            except KeystoreError:
                _invalid_key_message(key_name, keystore)

            if signer is None:
                generate_new_keys = skip_prompt is True or click.confirm(
                    "Generate new keys?"
                )
                if not generate_new_keys:
                    if click.confirm("Reuse existing key?"):
                        reused_key_name = input(
                            "Enter name of an existing keystore file: "
                        )
                        Path(keystore, key_name).write_bytes(
                            Path(keystore, reused_key_name).read_bytes()
                        )
                        Path(keystore, key_name + ".pub").write_bytes(
                            Path(keystore, reused_key_name + ".pub").read_bytes()
                        )
                    else:
                        raise KeystoreError(f"Could not load {key_name}")
                else:
                    break
    if generate_new_keys:
        if keystore is not None and (
            skip_prompt or click.confirm("Write keys to keystore files?")
        ):
            if password is None and not skip_prompt:
                password = input(
                    "Enter keystore password and press ENTER (can be left empty)"
                )
            private_pem = generate_and_write_rsa_keypair(
                path=Path(keystore, key_name), key_size=length, password=password
            )
            signer = load_signer_from_pem(private_pem)
        else:
            _, private_pem = generate_rsa_keypair(key_size=length)
            print(f"{role_name} key:\n\n{private_pem.decode()}\n\n")
            signer = load_signer_from_pem(private_pem)

    if signer is not None:
        return signer, _get_legacy_keyid(signer.public_key)
    raise KeystoreError(f"Could not load signer {key_name}")


def _setup_yubikey(
    auth_repo: AuthenticationRepository,
    role_name: str,
    key_name: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    certs_dir: Optional[Union[Path, str]] = None,
    key_size: int = 2048,
    yubikeys_to_skip: Optional[List] = None,
) -> Tuple[Dict, str]:
    print(f"Registering keys for {key_name}")
    taf_logger.debug(
        f"Starting YubiKey setup for role '{role_name}', key '{key_name}'. Checking if user wants to reuse an existing key."
    )
    while True:
        use_existing = click.confirm("Do you want to reuse already set up Yubikey?")
        taf_logger.debug(
            f"User chose to {'reuse' if use_existing else 'generate a new'} YubiKey for '{key_name}'."
        )

        if not use_existing:
            if not click.confirm(
                "WARNING - this will delete everything from the inserted key. Proceed?"
            ):
                if click.confirm("Cancel?"):
                    raise YubikeyError("Yubikey setup canceled")
                continue
        yubikeys = yk.yubikey_prompt(
            [key_name],
            pin_manager=auth_repo.pin_manager,
            role=role_name,
            taf_repo=auth_repo,
            registering_new_key=True,
            creating_new_key=not use_existing,
            pin_confirm=True,
            pin_repeat=True,
            yubikeys_to_skip=yubikeys_to_skip,
        )
        if yubikeys is not None:
            key, serial_num, key_name = yubikeys[0]
            if not use_existing:
                key = yk.setup_new_yubikey(
                    auth_repo.pin_manager, serial_num, scheme, key_size=key_size
                )

            if certs_dir is not None:
                # check if already exported
                if len(auth_repo.yubikey_store.get_roles_of_key(serial_num)) == 1:
                    # this is the first time that this key is being used (can only be used once per role)
                    yk.export_yk_certificate(certs_dir, key, serial=serial_num)
            return key, serial_num


def _load_remaining_keys_of_role(
    auth_repo: AuthenticationRepository,
    role: Role,
    loaded_keys_num: int,
    users_yubikeys_details: UserKeyData,
    yk_with_public_key: Dict,
    yubikey_keys,
    signers: List,
):
    """
    If a yubikey's public key was specified, meaning that it can be added as a
    verification key without being inserted, but the total number of signing
    keys is smaller than the threshold
    """
    taf_logger.debug(
        f"{loaded_keys_num} key(s) loaded so far for role '{role.name}' (threshold {role.threshold}). "
        f"Attempting to load more with known YubiKeys."
    )
    while loaded_keys_num < role.threshold:
        loaded_keys = []
        for key_name, public_key in yk_with_public_key.items():
            serial_num = _load_and_verify_yubikey(
                role.name,
                key_name,
                public_key,
                taf_repo=auth_repo,
            )
            if serial_num:
                loaded_keys_num += 1
                loaded_keys.append(key_name)
                signer = _create_signer(auth_repo, public_key, serial_num, key_name)
                signers.append(signer)
                yubikey_keys.remove(signer.public_key)

        if loaded_keys_num < role.threshold:
            if not click.confirm(
                f"Threshold of signing keys of role {role.name} not reached. Continue?"
            ):
                raise SigningError("Not enough signing keys")
            for key_name in loaded_keys:
                yk_with_public_key.pop(key_name)


def _load_and_verify_yubikey(
    role_name: str,
    key_name: str,
    public_key,
    taf_repo: AuthenticationRepository,
) -> Optional[str]:
    if not click.confirm(f"Sign using {key_name} Yubikey?"):
        return None
    while True:
        yubikeys = yk.yubikey_prompt(
            [key_name],
            role=role_name,
            pin_manager=taf_repo.pin_manager,
            taf_repo=taf_repo,
            registering_new_key=True,
            creating_new_key=False,
            pin_confirm=True,
            pin_repeat=True,
        )
        if yubikeys:
            yubikey = yubikeys[0]
            yk_pub_key_id = yubikey[0].keyid
            if yk_pub_key_id != public_key.keyid:
                print(
                    "Public key of the inserted key is not equal to the specified one."
                )
                if not click.confirm("Try again?"):
                    return None
            return yubikey[1]



import datetime
import secrets
from contextlib import contextmanager

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from taf.tuf.keys import load_signer_from_file

VALID_PIN = "123456"
WRONG_PIN = "111111"

INSERTED_YUBIKEY = None


class FakeYubiKey:
    def __init__(self, priv_key_path, pub_key_path, scheme, serial=None, pin=VALID_PIN):
        self.priv_key_pem = priv_key_path.read_bytes()
        self.pub_key_pem = pub_key_path.read_bytes()

        self._serial = serial if serial else secrets.randbelow(900000) + 100000
        self._pin = pin

        self.scheme = scheme
        self.priv_key = serialization.load_pem_private_key(
            self.priv_key_pem, None, default_backend()
        )
        self.pub_key = serialization.load_pem_public_key(
            self.pub_key_pem, default_backend()
        )

        self.tuf_key = load_signer_from_file(priv_key_path)

    @property
    def driver(self):
        return self

    @property
    def pin(self):
        return self._pin

    @property
    def serial(self):
        return self._serial

    def insert(self):
        """Insert YubiKey in USB slot."""
        global INSERTED_YUBIKEY
        INSERTED_YUBIKEY = self

    def is_inserted(self):
        """Check if YubiKey is in USB slot."""
        global INSERTED_YUBIKEY
        return INSERTED_YUBIKEY is self

    def remove(self):
        """Removes YubiKey from USB slot."""
        global INSERTED_YUBIKEY
        if INSERTED_YUBIKEY is self:
            INSERTED_YUBIKEY = None


class FakePivController:
    def __init__(self, driver):
        self._driver = driver

    @property
    def driver(self):
        return None

    def authenticate(self, *args, **kwargs):
        pass

    def change_pin(self, *args, **kwargs):
        pass

    def change_puk(self, *args, **kwargs):
        pass

    def generate_self_signed_certificate(self, *args, **kwargs):
        pass

    def get_pin_tries(self):
        return 1

    def get_certificate(self, _slot):
        name = x509.Name(
            [x509.NameAttribute(x509.NameOID.COMMON_NAME, self.__class__.__name__)]
        )
        now = datetime.datetime.utcnow()

        return (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(self._driver.pub_key)
            .serial_number(self._driver.serial)
            .not_valid_before(now)
            .not_valid_after(now + datetime.timedelta(days=365))
            .sign(self._driver.priv_key, hashes.SHA256(), default_backend())
        )

    def reset(self):
        pass

    def set_pin_retries(self, *args, **kwargs):
        pass

    def sign(self, slot, key_type, data, hash, padding):
        """Sign data using the same function as TUF"""
        if isinstance(data, str):
            data = data.encode("utf-8")

        signature = self._driver.priv_key.sign(data, padding, hash)
        return signature

    def verify_pin(self, pin):
        if self._driver.pin != pin:
            from yubikit.piv import InvalidPinError

            raise InvalidPinError(0)


class TargetYubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(
            keystore_path / "targets", keystore_path / "targets.pub", scheme
        )


class Root1YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root1", keystore_path / "root1.pub", scheme)


class Root2YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root2", keystore_path / "root2.pub", scheme)


class Root3YubiKey(FakeYubiKey):
    def __init__(self, keystore_path, scheme):
        super().__init__(keystore_path / "root3", keystore_path / "root3.pub", scheme)


@contextmanager
def _yk_piv_ctrl_mock(serial=None):
    global INSERTED_YUBIKEY

    if INSERTED_YUBIKEY is None:
        raise ValueError("No YubiKey found with the given interface(s)")

    yield [(FakePivController(INSERTED_YUBIKEY), INSERTED_YUBIKEY.serial)]



from __future__ import annotations
import json
from typing import Iterator, List, Optional, Dict
import attrs

from taf.constants import DEFAULT_ROLE_SETUP_PARAMS
from taf.exceptions import RolesKeyDataConversionError
from taf.models.validators import (
    filepath_validator,
    integer_validator,
    optional_type_validator,
    public_key_validator,
    role_number_validator,
    role_paths_validator,
)
from attrs import validators


@attrs.frozen
class Commitish:
    hash: str
    tag: Optional[str] = None

    @property
    def value(self) -> str:
        return self.tag if self.tag else self.hash

    @classmethod
    def from_hash(cls, hash: Optional[Commitish | str]):
        """
        Initialize cls from `hash`.
        If `hash` is already `Commitish`, returns self
        """
        if hash is None:
            return None
        if isinstance(hash, str):
            return cls(hash)  # type: ignore
        return hash  # type: ignore

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, str):
            other = Commitish(other)
        return self.value == other.value

    def __hash__(self):
        return self.value.__hash__()

    def __str__(self):
        return self.value

    def __repr__(self):
        return f'Commitish("{self.value}")'

    def to_json(self):
        """Serialize as a plain string instead of an object."""
        return json.dumps(str(self))


@attrs.define
class UserKeyData:
    public: Optional[str] = attrs.field(validator=public_key_validator, default=None)
    scheme: Optional[str] = attrs.field(
        validator=optional_type_validator(str),
        default=DEFAULT_ROLE_SETUP_PARAMS["scheme"],
    )
    present: bool = attrs.field(default=True)


@attrs.define
class Role:
    name: str = attrs.field(kw_only=True)
    threshold: int = attrs.field(
        validator=integer_validator, default=DEFAULT_ROLE_SETUP_PARAMS["threshold"]
    )
    scheme: Optional[str] = attrs.field(
        validator=optional_type_validator(str),
        default=DEFAULT_ROLE_SETUP_PARAMS["scheme"],
    )
    length: Optional[int] = attrs.field(
        validator=optional_type_validator(int),
        default=DEFAULT_ROLE_SETUP_PARAMS["length"],
    )
    yubikeys: Optional[List[str]] = attrs.field(default=None)
    number: int = attrs.field(validator=role_number_validator)
    # TODO remove after reworking add role/storing yubikey ids in the repo
    yubikey: Optional[bool] = attrs.field(default=None)

    @number.default  # type: ignore
    def _number_factory(self):
        return (
            len(self.yubikeys) if self.yubikeys else DEFAULT_ROLE_SETUP_PARAMS["number"]
        )

    @property
    def is_yubikey(self):
        return bool(
            self.yubikey is True or (self.yubikeys is not None and len(self.yubikeys))
        )

    @property
    def yubikey_ids(self):
        if not self.is_yubikey:
            return None
        return self.yubikeys


@attrs.define
class RootRole(Role):
    name: str = "root"


@attrs.define
class TargetsRole(Role):
    name: str = "targets"
    parent: Optional[Role] = attrs.field(default=None, kw_only=True)
    paths: Optional[List[str]] = attrs.field(
        kw_only=True, default=None, validator=role_paths_validator
    )
    terminating: Optional[bool] = attrs.field(
        validator=optional_type_validator(bool),
        default=DEFAULT_ROLE_SETUP_PARAMS["terminating"],
    )
    delegations: Optional[Dict[str, TargetsRole]] = attrs.field(
        kw_only=True,
        default={},
    )  # type: ignore

    def __attrs_post_init__(self):
        def _update_delegations(role):
            for role_name, delegated_role in role.delegations.items():
                delegated_role.name = role_name
                delegated_role.parent = role
                _update_delegations(delegated_role)

        if self.delegations:
            _update_delegations(self)


@attrs.define
class SnapshotRole(Role):
    name: str = "snapshot"


@attrs.define
class TimestampRole(Role):
    name: str = "timestamp"


@attrs.define
class MainRoles:
    root: RootRole
    targets: TargetsRole
    snapshot: SnapshotRole
    timestamp: TimestampRole


@attrs.define
class RolesKeysData:
    roles: MainRoles
    keystore: Optional[str] = attrs.field(validator=filepath_validator, default=None)  # type: ignore
    yubikeys: Optional[Dict[str, UserKeyData]] = attrs.field(default=None)

    def __attrs_post_init__(self):
        for role in RolesIterator(self.roles):
            if role.yubikeys:
                if not self.yubikeys:
                    raise RolesKeyDataConversionError(
                        exceptions=[
                            f"yubikeys of role {role.name} are listed, but yubikeys are not defined"
                        ]
                    )

                for key_id in role.yubikeys:
                    if key_id not in self.yubikeys:
                        raise RolesKeyDataConversionError(
                            exceptions=[
                                f"role {role.name} references yubikey {key_id}, but it is not specified"
                            ]
                        )


class RolesIterator:
    """
    Given an instance of MainRoles (which contains root, targets, snapshot or timestamp)
    or a targets (or delegated targets), iterate over all roles in the roles hierarchy.
    In case of MainRoles, iterate over root, targets, all delegated targets, snapshot and
    timestamp in that order. In case of a targets role, iterate over all of its nested
    targets roles
    """

    def __init__(
        self,
        roles: MainRoles | Role,
        include_delegations: Optional[bool] = True,
        skip_top_role: Optional[bool] = False,
    ):
        self.roles = roles
        self.include_delegations = include_delegations
        self.skip_top_role = skip_top_role

    def __iter__(self) -> Iterator:
        # Define the order of roles
        if isinstance(self.roles, MainRoles):
            roles = [
                self.roles.root,
                self.roles.targets,
                self.roles.snapshot,
                self.roles.timestamp,
            ]
        else:
            roles = [self.roles]

        def _dfs_delegations(role, skip_top_role=False):
            if not skip_top_role:
                yield role

            if self.include_delegations and hasattr(role, "delegations"):
                for delegation in role.delegations.values():
                    yield from _dfs_delegations(delegation)

        for role in roles:
            yield from _dfs_delegations(role, self.skip_top_role)


def compare_roles_data(old_data: RolesKeysData, new_data: RolesKeysData):
    added_roles = []
    removed_roles = []
    current_roles = {
        role.name: role
        for role in RolesIterator(old_data.roles.targets, skip_top_role=True)
    }
    new_roles = {
        role.name: role
        for role in RolesIterator(new_data.roles.targets, skip_top_role=True)
    }

    for role_name, role in current_roles.items():
        if role_name not in new_roles:
            removed_roles.append(role)

    for role_name, role in new_roles.items():
        if role_name not in current_roles:
            added_roles.append(role)

    return added_roles, removed_roles


@attrs.define(slots=True, frozen=True)
class KeyEntry:
    """
    Single entry inside keys-mapping.json.
    """

    public: str = attrs.field(validator=validators.instance_of(str))
    keyid: str = attrs.field(validator=validators.instance_of(str))
    scheme: str = attrs.field(validator=validators.instance_of(str))

    @staticmethod
    def from_dict(data: Dict[str, str]) -> "KeyEntry":
        """
        Parse one object (the value part in keys-mapping.json).
        """
        return KeyEntry(  # type: ignore[call-arg]
            public=data["public"],
            keyid=data["keyid"],
            scheme=data["scheme"],
        )


@attrs.define(slots=True, frozen=True)
class KeysMapping:
    """
    Wrapper around the whole keys-mapping.json file.
    Maps key-name -> KeyEntry.
    """

    entries: Dict[str, KeyEntry]

    @staticmethod
    def from_dict(mapping: Dict[str, Dict[str, str]]) -> "KeysMapping":
        return KeysMapping(  # type: ignore[call-arg]
            entries={name: KeyEntry.from_dict(info) for name, info in mapping.items()}
        )

    def find_name_by_public(self, public_pem: str) -> Optional[str]:
        for name, info in self.entries.items():
            if info.public == public_pem:
                return name
        return None

    def find_name_by_keyid(self, keyid: str) -> Optional[str]:
        for name, info in self.entries.items():
            if info.keyid == keyid:
                return name
        return None



from logging import DEBUG, ERROR
from typing import Dict, Optional
import click

from pathlib import Path
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError, YubikeyError

# from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.log import taf_logger
from taf.tuf.keys import get_sslib_key_from_value
from taf.tuf.repository import MAIN_ROLES
import taf.yubikey.yubikey as yk
from taf.yubikey.yubikey_manager import PinManager


@log_on_start(DEBUG, "Exporting public pem from YubiKey", logger=taf_logger)
@log_on_end(DEBUG, "Exported public pem from YubuKey", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while exporting public pem from YubiKey: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def export_yk_public_pem(
    path: Optional[str] = None, serial: Optional[str] = None
) -> None:
    """
    Export public key from a YubiKey and save it to a file or print to console.

    Arguments:
        path (optional): Path to a file to which the public key should be written.
        The key is printed to console if file path is not provided.

    Side Effects:
       Write public key to a file if path is specified

    Returns:
        None
    """
    try:
        serials = [serial] if serial else yk.get_serial_nums()

        if not len(serials):
            print("YubiKey not inserted.")
            return

        for serial in serials:
            pub_key = yk.get_piv_public_key_tuf(serial=serial)
            keyid = yk._get_legacy_keyid(pub_key)
            pub_key_pem = pub_key.keyval["public"]
            if path is None:
                print(f"Serial: {serial}")
                print(f"Key id: {keyid}")
                print(pub_key_pem)
            else:
                if not path.endswith(".pub"):
                    path = f"{path}.pub"
                pem_path = Path(path)
                parent = pem_path.parent
                parent.mkdir(parents=True, exist_ok=True)
                pem_path.write_text(pub_key_pem)
    except Exception:
        print("Could not export the public key. Check if a YubiKey is inserted")
        return


@log_on_start(DEBUG, "Exporting certificate from YubiKey", logger=taf_logger)
@log_on_end(DEBUG, "Exported certificate from YubuKey", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while exporting certificate from YubiKey: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def export_yk_certificate(
    path: Optional[str] = None, serial: Optional[str] = None
) -> None:
    """
    Export certificate from the YubiKey.

    Arguments:
        path (optional): Path to a file to which the certificate key should be written.
        Will be written to the user's home directory by default

    Side Effects:
       Write certificate to a file

    Returns:
        None
    """
    try:
        serials = [serial] if serial else yk.get_serial_nums()

        if not len(serials):
            print("YubiKey not inserted.")
            return
        for serial in serials:
            pub_key_pem = yk.export_piv_pub_key(serial=serial).decode("utf-8")
            scheme = DEFAULT_RSA_SIGNATURE_SCHEME
            key = get_sslib_key_from_value(pub_key_pem, scheme)
            yk.export_yk_certificate(path, key, serial)
    except Exception as e:
        print(e)
        print("Could not export certificate. Check if a YubiKey is inserted")
        return


@log_on_start(DEBUG, "Listing roles of inserted YubiKesy", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while listing roles of inserted YubiKeys: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def get_yk_roles(path: str, serial: Optional[str] = None) -> Dict:
    """
    List all roles that the inserted YubiKey whose metadata files can be signed by this YubiKey.
    In case of delegated targets roles, include the delegation paths.

    Arguments:
        path: Authentication repository's path.
    Side Effects:
        None

    Returns:
        A dictionary containing roles and delegated paths in case of delegated target roles
    """
    serials = [serial] if serial else yk.get_serial_nums()
    roles_per_yubikes: Dict = {}

    if not len(serials):
        print("YubiKey not inserted.")
        return roles_per_yubikes

    auth = AuthenticationRepository(path=path)
    for serial in serials:
        pub_key = yk.get_piv_public_key_tuf(serial=serial)
        roles = auth.find_associated_roles_of_key(pub_key)
        roles_with_paths: Dict = {role: {} for role in roles}
        for role in roles:
            if role not in MAIN_ROLES:
                roles_with_paths[role] = auth.get_role_paths(role)
        roles_per_yubikes[serial] = roles_with_paths
    return roles_per_yubikes


@log_on_start(DEBUG, "Setting up a new signing YubiKey", logger=taf_logger)
@log_on_end(DEBUG, "Finished setting up a new signing YubiKey", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while setting up a new YubiKey: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def setup_signing_yubikey(
    pin_manager: PinManager, certs_dir: Optional[str] = None, key_size: int = 2048
) -> None:
    """
    Delete everything from the inserted YubiKey, generate a new key and copy it to the YubiKey.
    Optionally export and save the certificate to a file.

    Arguments:
        certs_dir (optional): Path to a directory where the exported certificate should be stored.

    Side Effects:
       None

    Returns:
        None
    """
    if not click.confirm(
        "WARNING - this will delete everything from the inserted key. Proceed?"
    ):
        return
    yubikeys = yk.yubikey_prompt(
        ["new Yubikey"],
        pin_manager=pin_manager,
        creating_new_key=True,
        pin_confirm=True,
        pin_repeat=True,
        prompt_message="Please insert the new Yubikey and press ENTER",
    )
    if yubikeys:
        _, serial_num, _ = yubikeys[0]
        key = yk.setup_new_yubikey(pin_manager, serial_num, key_size=key_size)
        yk.export_yk_certificate(certs_dir, key, serial_num)
    else:
        raise YubikeyError("Could not generate a new key")


@log_on_start(DEBUG, "Setting up a new test YubiKey", logger=taf_logger)
@log_on_end(DEBUG, "Finished setting up a test YubiKey", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while setting up a test YubiKey: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
)
def setup_test_yubikey(
    pin_manager: PinManager,
    key_path: str,
    key_size: Optional[int] = 2048,
    serial: Optional[str] = None,
) -> None:
    """
    Reset the inserted yubikey, set default pin and copy the specified key
    to it.

    Arguments:
        key_path: Path to a key which should be copied to a YubiKey.

    Side Effects:
       None

    Returns:
        None
    """
    if serial is None:
        serials = yk.get_serial_nums()
        if not len(serials):
            raise YubikeyError("YubiKey not inserted")
        if len(serials) > 1:
            raise YubikeyError("Insert only one YubiKey")

    if not click.confirm("WARNING - this will reset the inserted key. Proceed?"):
        return

    serial = serials[0]
    key_pem_path = Path(key_path)
    key_pem = key_pem_path.read_bytes()

    print(f"Importing RSA private key from {key_path} to Yubikey...")
    pin = yk.DEFAULT_PIN
    pin_manager.add_pin(serial, pin)

    pub_key = yk.setup(
        pin, serial, "Test Yubikey", private_key_pem=key_pem, key_size=key_size
    )
    print("\nPrivate key successfully imported.\n")
    print("\nPublic key (PEM): \n{}".format(pub_key.decode("utf-8")))
    print("Pin: {}\n".format(pin))



import click
from taf.api.yubikey import (
    export_yk_certificate,
    export_yk_public_pem,
    get_yk_roles,
    setup_signing_yubikey,
    setup_test_yubikey,
)
from taf.exceptions import YubikeyError
from taf.repository_utils import find_valid_repository
from taf.tools.cli import catch_cli_exception
from taf.tools.repo import pin_managed
from taf.yubikey.yubikey import list_connected_yubikeys, list_all_devices


def check_pin_command():
    @click.command(help="Checks if the specified pin is valid")
    @click.option("--pin", default=None, help="PIN to be checked")
    @click.option("--serial", default=None, type=int, help="Serial number of a YubiKey. Has to be provided if more than one YK is inserted")
    @catch_cli_exception(handle=YubikeyError)
    def check_pin(pin, serial):
        try:
            from taf.yubikey.yubikey import is_valid_pin, get_and_validate_pin

            if serial is None and len(list_all_devices()) > 1:
                click.echo("More than one YubiKey is inserted. Please specify the serial number of the YubiKey to be used")
                return
            name = serial or "YubiKey"
            if pin is None:
                get_and_validate_pin(name, serial=serial)
            else:
                is_valid, retries = is_valid_pin(pin, serial=serial)
                if not is_valid:
                    click.echo(f"Invalid PIN. You have {retries} retries left")
        except Exception as e:
            click.echo(f"Error: {e}")
            return
    return check_pin


def export_pub_key_command():
    @click.command(
        help="Export public keys of the inserted YubiKeys"
    )
    @click.option(
        "--output",
        help="File to which the exported public key will be written. The result will be written to the console if path is not specified",
    )
    @catch_cli_exception(handle=YubikeyError)
    def export_pub_key(output):
        export_yk_public_pem(output)

    return export_pub_key


def get_roles_command():
    @click.command(
        help="List roles the inserted YubiKey is allowed to sign."
    )
    @catch_cli_exception(handle=YubikeyError, print_error=True)
    @click.option(
        "--path",
        default=".",
        help="Authentication repository's location. If not specified, set to the current directory",
    )
    @catch_cli_exception(handle=YubikeyError)
    def get_roles(path):
        path = find_valid_repository(path)
        roles_with_paths = get_yk_roles(path)
        for serial, roles_and_paths in roles_with_paths.items():
            print(f"\nSerial: {serial}")
            for role, paths in roles_and_paths.items():
                print(f"\n{role}")
                for path in paths:
                    print(f"\n -{path}")

    return get_roles


def export_certificate_command():
    @click.command(
        help="Export certificates of the inserted YubiKeys"
    )
    @click.option(
        "--output",
        help="File to which the exported certificate key will be written. The result will be written to the user's home directory by default",
    )
    @catch_cli_exception(handle=YubikeyError)
    def export_certificate(output):
        export_yk_certificate(output)

    return export_certificate


def list_key_command():
    @click.command(help="List All Connected Keys and their information")
    @catch_cli_exception(handle=YubikeyError)
    def list_keys():
        list_connected_yubikeys()
    return list_keys


def setup_signing_key_command():
    @click.command(
        help="""Generate a new key on the yubikey and set the pin. Export the generated certificate
        to the specified directory.
        WARNING - this will delete everything from the inserted key."""
    )
    @click.option(
        "--certs-dir",
        help="Path of the directory where the exported certificate will be saved. Set to the user home directory by default",
    )
    @catch_cli_exception(handle=YubikeyError)
    @pin_managed
    def setup_signing_key(certs_dir, pin_manager):
        setup_signing_yubikey(pin_manager, certs_dir, key_size=2048)

    return setup_signing_key


def setup_test_key_command():
    @click.command(
        help="""Copies the specified key onto the inserted YubiKey
        WARNING - this will reset the inserted key."""
    )
    @click.argument("key-path")
    @catch_cli_exception(handle=YubikeyError)
    @pin_managed
    def setup_test_key(key_path, pin_manager):
        setup_test_yubikey(pin_manager, key_path)

    return setup_test_key


def attach_to_group(group):
    group.add_command(check_pin_command(), name='check-pin')
    group.add_command(export_pub_key_command(), name='export-pub-key')
    group.add_command(get_roles_command(), name='get-roles')
    group.add_command(export_certificate_command(), name='export-certificate')
    group.add_command(list_key_command(), name='list-key')
    group.add_command(setup_signing_key_command(), name='setup-signing-key')
    group.add_command(setup_test_key_command(), name='setup-test-key')
